/* based on: */

/*	$OpenBSD: inet_ntop.c,v 1.8 2008/12/09 19:38:38 otto Exp $	*/

/* Copyright (c) 1996 by Internet Software Consortium.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND INTERNET SOFTWARE CONSORTIUM DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL INTERNET SOFTWARE
 * CONSORTIUM BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

/*
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: sm_inet_ipv62a.c,v 1.6 2014/01/30 16:37:29 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/ctype.h"
#include "sm/net.h"
#include "sm/io.h"
#include "sm/str.h"
#include "sm/nameser.h"	/* some INADDR constants */

#if MTA_NETINET6
/*
**  SM_INET_IPV6STR -- convert an IPv6 address to an ASCII string
**
**	Parameters:
**		ipv6 -- IPv6 address
**		dst -- append (compressed) ASCII representation of address (output)
**
**	Returns:
**		error: ENOSPC
*/

sm_ret_T
sm_inet_ipv6str(ipv6_P ipv6, sm_str_P dst)
{
	uchar *src;
	int i;
	struct { int base, len; } best, cur;
	uint words[IN6ADDRSZ / INT16SZ];

	/*
	 * Preprocess:
	 *	Copy the input (bytewise) array into a wordwise array.
	 *	Find the longest run of 0x00's in src[] for :: shorthanding.
	 */
	src = (uchar *) ipv6;
	memset(words, '\0', sizeof words);
	for (i = 0; i < IN6ADDRSZ; i++)
		words[i / 2] |= (src[i] << ((1 - (i % 2)) << 3));
	best.base = -1;
	cur.base = -1;
	for (i = 0; i < (IN6ADDRSZ / INT16SZ); i++) {
		if (words[i] == 0) {
			if (cur.base == -1)
				cur.base = i, cur.len = 1;
			else
				cur.len++;
		} else {
			if (cur.base != -1) {
				if (best.base == -1 || cur.len > best.len)
					best = cur;
				cur.base = -1;
			}
		}
	}
	if (cur.base != -1) {
		if (best.base == -1 || cur.len > best.len)
			best = cur;
	}
	if (best.base != -1 && best.len < 2)
		best.base = -1;

	for (i = 0; i < (IN6ADDRSZ / INT16SZ); i++) {
		/* Are we inside the best run of 0x00's? */
		if (best.base != -1 && i >= best.base &&
		    i < (best.base + best.len)) {
			if (i == best.base)
				SM_STR_PUT(dst, ':');
			continue;
		}
		/* Are we following an initial run of 0x00s or any real hex? */
		if (i != 0)
			SM_STR_PUT(dst, ':');

		/* Is this address an encapsulated IPv4? */
		if (i == 6 && best.base == 0 &&
		    (best.len == 6 || (best.len == 5 && words[5] == 0xffff)))
		{
			sm_ret_T ret;

			/* use src+12 to avoid access to struct elements? ? */
			ret = sm_inet_ipv4str(ipv6->sm6_addr.sm6_addr32[3], dst);
			if (sm_is_err(ret))
				return ret;
			break;
		}
		sm_strprintf(dst, "%x", words[i]);
	}
	/* Was it a trailing run of 0x00's? */
	if (best.base != -1 && (best.base + best.len) == (IN6ADDRSZ / INT16SZ))
		SM_STR_PUT(dst, ':');

	return SM_SUCCESS;
}

/*
**  SM_INET_IN6ADDR2STR -- convert an IPv6 address struct to an ASCII string
**
**	Parameters:
**		in6addr -- IPv6 address struct
**		dst -- append ASCII representation of address (output)
**
**	Returns:
**		error: ENOSPC
*/

sm_ret_T
sm_inet_in6addr2str(struct in6_addr in6addr, sm_str_P dst)
{
	uchar *src;
	ipv6_T ipv6;

	src = (uchar *)&in6addr;
	memcpy(&ipv6, src, sizeof(ipv6));
	return sm_inet_ipv6str(&ipv6, dst);
}
#endif /* MTA_NETINET6 */
